﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Net;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;
using TSTunnels.Common;
using System.Diagnostics;
using System.Net.Sockets;

namespace TSTunnels.Server
{
	public partial class Server : Form
	{
		public Server()
		{
			InitializeComponent();
		}

		private IntPtr mHandle = IntPtr.Zero;
		private BinaryReader reader;
		private void Server_Load(object sender, EventArgs e)
		{
			mHandle = WtsApi32.WTSVirtualChannelOpen(IntPtr.Zero, -1, ChannelMessage.ChannelName);
			if (mHandle == IntPtr.Zero)
			{
				textBox2.SelectedText = new Win32Exception() + Environment.NewLine;
				return;
			}

			int len;
			IntPtr buffer;
			var b = WtsApi32.WTSVirtualChannelQuery(mHandle, WtsVirtualClass.WTSVirtualFileHandle, out buffer, out len);
			if (!b)
			{
				textBox2.Text = new Win32Exception().ToString();
				return;
			}
			var fileHandle = new SafeFileHandle(Marshal.ReadIntPtr(buffer), true);
			WtsApi32.WTSFreeMemory(buffer);

			reader = new BinaryReader(new BufferedStream(new FileStream(fileHandle, FileAccess.ReadWrite, 0x1000, true)));
			backgroundWorker1.RunWorkerAsync();

			timer1.Enabled = true;
		}

		private void Server_FormClosed(object sender, FormClosedEventArgs e)
		{
			if (reader != null) reader.Close();
			var ret = WtsApi32.WTSVirtualChannelClose(mHandle);
			backgroundWorker1.CancelAsync();
		}

		private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
		{
			while (!backgroundWorker1.CancellationPending)
			{
				try
				{
					var len = reader.ReadInt32();
					var buff = reader.ReadBytes(len);
					backgroundWorker1.ReportProgress(0, buff);
				}
				catch (OperationCanceledException)
				{
					return;
				}
			}
		}

		private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			if (backgroundWorker1.CancellationPending) return;
			MessageReceived(ChannelMessage.FromByteArray((byte[])e.UserState));
		}

		private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
		{
			var r = e.Result;
		}

		private void WriteMessage(ChannelMessage message)
		{
			var data = message.ToByteArray();
			int written;
			var ret = WtsApi32.WTSVirtualChannelWrite(mHandle, data, data.Length, out written);
			if (ret) return;
			textBox2.SelectedText = new Win32Exception() + Environment.NewLine;
			return;
		}

		private int connectionCount;
		private readonly Dictionary<int, Stream> streams = new Dictionary<int, Stream>();

		private void CreateLocalPort(string listenAddress, int listenPort, string connectAddress, int connectPort)
		{
			var listener = new TcpListener(IPAddress.Parse(listenAddress), listenPort);
			listener.Start();
			listener.BeginAcceptTcpClient(listener_AcceptTcpClient, (AcceptTcpClientCallback)delegate(IAsyncResult ar)
			{
				var streamIndex = ++connectionCount;
				streams[streamIndex] = listener.EndAcceptTcpClient(ar).GetStream();
				WriteMessage(new ConnectRequest(streamIndex, connectAddress, connectPort));
				return listener;
			});
		}

		private delegate TcpListener AcceptTcpClientCallback(IAsyncResult ar);
		private static void listener_AcceptTcpClient(IAsyncResult ar)
		{
			var callback = (AcceptTcpClientCallback)ar.AsyncState;
			var listener = callback(ar);
			listener.BeginAcceptTcpClient(listener_AcceptTcpClient, callback);
		}

		private readonly Dictionary<int, Action<int>> remoteServers = new Dictionary<int, Action<int>>();
		private void CreateRemotePort(string listenAddress, int listenPort, string connectAddress, int connectPort)
		{
			var listenIndex = ++connectionCount;
			remoteServers[listenIndex] = index =>
			{
				var client = new TcpClient();
				client.BeginConnect(connectAddress, connectPort, client_Connect, new KeyValuePair<int, TcpClient>(index, client));
			};
			WriteMessage(new ListenRequest(listenIndex, listenAddress, listenPort));
		}

		private void client_Connect(IAsyncResult ar)
		{
			var state = (KeyValuePair<int, TcpClient>)ar.AsyncState;
			var streamIndex = state.Key;
			var tcpClient = state.Value;
			try
			{
				tcpClient.EndConnect(ar);
				var stream = streams[streamIndex] = tcpClient.GetStream();
				new StreamPump(stream,
					data => WriteMessage(new StreamData(streamIndex, data)),
					ex => WriteMessage(new StreamError(streamIndex, ex.ToString()))).Pump();
			}
			catch (Exception ex)
			{
				WriteMessage(new StreamError(streamIndex, ex.ToString()));
				Debug.Print(ex.ToString());
			}
		}

		private bool portsCreated;
		public void MessageReceived(ChannelMessage msg)
		{
			switch (msg.Type)
			{
				case MessageType.HelloReply:
					{
						timer1.Enabled = false;
						textBox2.SelectedText = msg + Environment.NewLine;
						if (!portsCreated)
						{
							portsCreated = true;
							CreateLocalPort("127.0.0.1", 9000, "darkskies.za.net", 22);
							CreateRemotePort("127.0.0.1", 9001, "darkskies.za.net", 22);
						}
					}
					break;
				case MessageType.ConnectResult:
					{
						var result = (ConnectResult)msg;
						var stream = streams[result.StreamIndex];
						new StreamPump(stream,
							data => WriteMessage(new StreamData(result.StreamIndex, data)),
							ex => WriteMessage(new StreamError(result.StreamIndex, ex.ToString()))).Pump();
					}
					break;
				case MessageType.ListenResponse:
					{
						var result = (ListenResponse)msg;
						var server = remoteServers[result.ListenIndex];
						server(result.StreamIndex);
					}
					break;
				case MessageType.StreamData:
					{
						var data = (StreamData)msg;
						var stream = streams[data.StreamIndex];
						try
						{
							stream.Write(data.Data, 0, data.Data.Length);
						}
						catch (Exception ex)
						{
							WriteMessage(new StreamError(data.StreamIndex, ex.ToString()));
						}
					}
					break;
				case MessageType.StreamError:
					{
						var error = (StreamError)msg;
						var stream = streams[error.StreamIndex];
						if (error.Exception.Length != 0)
						{
							try
							{
								var data = Encoding.UTF8.GetBytes(error.Exception);
								stream.Write(data, 0, data.Length);
							}
							catch (IOException)
							{
							}
							return;
						}
						stream.Close();
						streams.Remove(error.StreamIndex);
					}
					break;
				default:
					{
						textBox2.SelectedText = msg + Environment.NewLine;
					}
					break;
			}
		}

		private void button1_Click(object sender, EventArgs e)
		{
			WriteMessage(new HelloRequest());
		}

		private void timer1_Tick(object sender, EventArgs e)
		{
			WriteMessage(new HelloRequest());
		}
	}
}
